package deltix.samples.test;

import deltix.qsrv.hf.pub.*;
import deltix.qsrv.hf.pub.codec.*;
import deltix.qsrv.hf.pub.md.*;
import deltix.qsrv.hf.tickdb.pub.*;

import deltix.util.memory.*;
import org.junit.*;
import static org.junit.Assert.*;

/**
 *  Store, then read messages with a large number of fields. This also tests
 *  the transmission of large unicode symbols as part of the metadata.
 *  Finally, we test that the compiled codec factory falls back to
 *  interpreting logic when the class is too big.
 */

public class Test_HugeClass extends TDBTestBase {
    public static final String      STREAM_KEY = "test.stream";
    public static final int         NUM_MESSAGES = 20;

    static class HugeClassTester {
        private final int                   numFields;
        private final RecordClassDescriptor cd;
        private final CodecFactory          codecFactory;

        public HugeClassTester (int numFields, CodecFactory codecFactory) {
            this.codecFactory = codecFactory;
            this.numFields = numFields;

            DataField []    fields = new DataField [numFields];

            for (int ii = 0; ii < numFields; ii++) {
                fields [ii] =
                    new NonStaticDataField (
                        "\u041F\u043E\u043B\u0435" + ii,
                        "\u0412\u0441\u0435\u043C \u0431\u043E\u043B\u044C\u0448\u043E\u0439 \u043F\u0440\u0438\u0432\u0435\u0442" + ii,
                        new IntegerDataType ("INT8", true),
                        null
                    );
            }

            cd = new RecordClassDescriptor ("MyClass", "Title", false, null, fields);
        }

        public void                 run (DXTickDB db) throws Exception {
            StreamOptions               options =
                new StreamOptions (
                    StreamScope.DURABLE,
                    "Stream Name",
                    "Description Line1\nLine 2\nLine 3",
                    StreamOptions.MAX_DISTRIBUTION
                );

            options.setFixedType (cd);

            deleteIfExists(db, STREAM_KEY);
            final DXTickStream      stream = db.createStream (STREAM_KEY, options);

            RecordClassDescriptor   classDescriptor = stream.getFixedType ();
            RawMessage              msg = new RawMessage (classDescriptor);

            msg.setSymbol("DLTX");

            LoadingOptions          loptions = new LoadingOptions (true);
            TickLoader              loader = stream.createLoader (loptions);
            MemoryDataOutput        dataOutput = new MemoryDataOutput ();            

            FixedUnboundEncoder     encoder =
                codecFactory.createFixedUnboundEncoder (classDescriptor);

            for (int ii = 0; ii < NUM_MESSAGES; ii++) {
                dataOutput.reset ();
                encoder.beginWrite (dataOutput);

                int fidx;

                for (fidx = 0; encoder.nextField (); fidx++) {
                    int         x = (ii + fidx) & 0xFF;

                    if (x == 128)
                        encoder.writeNull ();
                    else {
                        if (x > 128)
                            x -= 256;
                        
                        encoder.writeInt (x);
                    }
                }
                
                assertEquals (numFields, fidx);

                msg.setBytes (dataOutput, 0);

                loader.send (msg);
            }

            loader.close ();
            
            SelectionOptions        soptions = new SelectionOptions (true, false);
            TickCursor              cursor = stream.select (0, soptions);
            MemoryDataInput         in = new MemoryDataInput ();
            UnboundDecoder          decoder =
                codecFactory.createFixedUnboundDecoder (classDescriptor);


            int                     midx;
            
            for (midx = 0; cursor.next (); midx++) {
                msg = (RawMessage) cursor.getMessage ();

                in.setBytes (msg.data, msg.offset, msg.length);
                decoder.beginRead (in);

                int fidx;

                for (fidx = 0; decoder.nextField (); fidx++) {
                    int         x = (midx + fidx) & 0xFF;

                    if (x == 128)
                        assertTrue (decoder.isNull ());
                    else {
                        assertFalse (decoder.isNull ());
                        
                        if (x > 128)
                            x -= 256;

                        assertEquals (x, decoder.getInt ());
                    }
                }

                assertEquals (numFields, fidx);
            }

            cursor.close ();            
        }
    }

    @Test
    public void             compFallbackLocal () throws Exception {
        //  Test fallback that turns off compilation of a huge class
        new HugeClassTester(1000, CodecFactory.COMPILED).run (getTickDb());
    }

}
